/*
 * Copyright (C) 2013 Kaan Onarlioglu <onarliog@ccs.neu.edu>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * Kaan Onarlioglu -- www.onarlioglu.com
 * NEU SecLab      -- seclab.ccs.neu.edu
 *
 * main.c, ver.2013.10.13
 *
 * Run-PrivExec.
 */

#include <stdio.h>
#include <stdlib.h>

#include "privexec.h"

/*
 * We have a simple checkpoint & rollback system to reverse all the changes made
 * to the system after PrivExec application exits, or after an error during the
 * launch.
 */

/* Checkpoints for rollback. */
#define THE_END             8
#define MOUNTING_SYS        7
#define MOUNTING_OVERLAYS   6
#define MOUNTING_CONTAINERS 5
#define KEY_INSERTED        4
#define CREATING_DIRS       3
#define GENERATING_NAMES    2
#define START               1

static unsigned checkpoint;

static inline void set_checkpoint( unsigned p) {
    checkpoint = p;
}

static void _rollback( unsigned checkpoint) {

    /* Fall through all the way. */
    switch ( checkpoint) {
    case THE_END:
        raise_privileges();
    case MOUNTING_SYS:
        umount_sys();
    case MOUNTING_OVERLAYS:
        umount_overlays();
    case MOUNTING_CONTAINERS:
        umount_containers();
        delete_enc_dirs();
    case KEY_INSERTED:
        drop_privileges();
        remove_key();
    case CREATING_DIRS:
        remove_dirs();
    case GENERATING_NAMES:
        free_memory(); /* We are about to die, but still... */
    case START:
        break;
    default:
        print_error( "Error trying to rollback. Seriously?\n");
        exit( EXIT_FAILURE);
    }
}

static void rollback() {
    _rollback( checkpoint);
}

int main( int argc, char *argv[]) {

    /* Pass the executable name and arguments. */
    if ( argc < 2) {
        printf( "Usage: %s <executable> <arguments...>\n", argv[0]);
        return EXIT_FAILURE;
    }

    set_cleanup_func( rollback); /* Register the rollback function */
    set_checkpoint( START);

    /* 0 - Initialize. */
    init();

    /*
     * We don't want to be able to create directories in locations we are not
     * supposed to access.
     */
    drop_privileges();

    /* 1 - Generate directory names for this PrivExec session. */
    set_checkpoint( GENERATING_NAMES);
    generate_dir_names();

    /* 2 - Create the directories. */
    set_checkpoint( CREATING_DIRS);
    create_dirs();

    /* 3 - Insert a dummy ecryptfs key into the keyring. */
    insert_key();
    set_checkpoint( KEY_INSERTED);

    /*
     * 4 - Time to make this process private. Secure containers MUST be mounted
     * by the private process itself so that the kernel knows who owns them.
     */
    fork_privexec();

    /*
     * After this point we need privileges to mount filesystems and later do a
     * chroot. Use Linux capabilities, or set this binary suid root, or use
     * sudo. We will drop everything before launching the PrivExec application.
     */
    raise_privileges();

    /* 5 - Mount the secure containers. */
    set_checkpoint( MOUNTING_CONTAINERS);
    mount_containers();
    remove_key(); /* Don't need it anymore. */

    /* 6 - Mount the secure containers. */
    set_checkpoint( MOUNTING_OVERLAYS);
    mount_overlays();

    /* 7 - Bind-mount special filesystems. */
    set_checkpoint( MOUNTING_SYS);
    mount_sys();

    /* Setup complete. */
    drop_privileges();
    set_checkpoint( THE_END);

    /* 8 - Start Privexec! */
    print_green( "PrivExec starting!\n");
    run_privexec( argv + 1);
    print_red( "PrivExec exiting. Bye!\n");

    /* 9 - Parent will clean up. */
    rollback();

    return EXIT_SUCCESS;
}
