/*
 * Copyright (C) 2013 Kaan Onarlioglu <onarliog@ccs.neu.edu>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * Kaan Onarlioglu -- www.onarlioglu.com
 * NEU SecLab      -- seclab.ccs.neu.edu
 *
 * privexec.c, ver.2013.10.13
 *
 * Functions to set up the environment for running PrivExec.
 */

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/mount.h>
#include <sys/syscall.h>
#include <ecryptfs.h>

#include "privexec.h"

/*
 * Kernel's PrivExec clone flag. Let's not install & include the new kernel
 * headers for now; this flag is all we need.
 */
#define CLONE_PRIVEXEC 0x02000000


/*
 * The Evil Globals -- because we need them for proper rollback in case
 * something goes wrong.
 */

/* Mount points. */
static char **mnt_points;
static unsigned mnt_points_len;
static char **mnt_targets;
static unsigned mnt_targets_len;
static char **mnt_sys;
static unsigned mnt_sys_len;
#define MOUNT_BUF_LEN 3 /* Malloc size for mount point buffers. */

/* Directories. */
static char *tmp_dir;
static char *storage_dir;
static char *union_dir;
static char **dec_dirs;
static unsigned dec_dirs_len;
static char **enc_dirs;
static unsigned enc_dirs_len;
static char **dirs;

/* A dummy ecryptfs key for mounting the secure containers. */
static char dummy_key[ ECRYPTFS_SIG_SIZE_HEX + 1];


/*
 * Internal functions.
 */

/*
 * Reads currently mounted filesystems from /proc/mounts. Returns in a string
 * array, and its length in an out argument.
 */
static char **read_mounts_from_proc( unsigned *mount_proc_len) {

    char *buf;
    char *tok_buf;
    char *mnt;
    char **mnt_proc;
    unsigned cur;
    unsigned max;
    unsigned file_len;

    /* Read the file and fix the buffer. */
    buf = read_text_file("/proc/mounts", &file_len);
    buf[ file_len - 1] = '\0'; /* Why is proc doing this? */

    /*
     * Parse mounted filesystems. We are only interested in mount targets -- the
     * second field in every line.
     */
    tok_buf = buf;
    max = MOUNT_BUF_LEN;
    cur = 0;
    mnt_proc = malloc( max * sizeof( *mnt_proc));

    while ( strsep( &tok_buf, " ")) { /* Skip first token in each line. */
        mnt = strsep( &tok_buf, " ");
        do_strdup( mnt_proc + cur, mnt);

        ++cur;
        mnt_proc = try_realloc( mnt_proc, cur, &max,
                                MOUNT_BUF_LEN, sizeof( *mnt_proc));

        strsep( &tok_buf, "\n"); /* Skip to the end of line. */
    }
    free( buf);

    if ( mount_proc_len)
        *mount_proc_len = cur;
    return mnt_proc;
}

/*
 * Reads a PrivExec launcher config file and returns a list containing the
 * listed mount points, and the list's size in an out argument.
 */
static char **read_mounts_from_conf( unsigned *mnt_conf_len, char *path) {

    char *buf;
    char *tok_buf;
    char *mnt;
    char **mnt_conf;
    unsigned cur;
    unsigned max;

    /* Read config file. */
    buf = read_text_file( path, NULL);

    /* Parse lines. Every line contains one entry. */
    tok_buf = buf;
    max = MOUNT_BUF_LEN;
    cur = 0;
    mnt_conf = malloc( max * sizeof( *mnt_conf));

    while( ( mnt = strsep( &tok_buf, "\n"))) {
        if ( *trim( mnt) != '\0') {
            do_strdup( mnt_conf + cur, mnt);
            ++cur;
            mnt_conf = try_realloc( mnt_conf, cur, &max,
                                    MOUNT_BUF_LEN, sizeof( *mnt_conf));
        }
    }
    free( buf);

    if ( mnt_conf_len)
        *mnt_conf_len = cur;
    return mnt_conf;
}

/*
 * Reads the mount points specified in the PrivExec config file, reads the
 * currently mounted filesystems, and takes their intersection to determine the
 * encrypted mount points PrivExec needs to set up.
 */
static void get_mount_points() {

    char **mnt_proc;
    char **mnt_conf;
    unsigned mnt_proc_len;
    unsigned mnt_conf_len;
    unsigned max;
    int i;
    int j;

    mnt_proc = read_mounts_from_proc( &mnt_proc_len);
    mnt_conf = read_mounts_from_conf( &mnt_conf_len, PRIVEXEC_USR_MNT_CONF);

    max = MOUNT_BUF_LEN;
    mnt_points_len = 0;
    mnt_points = malloc( max * sizeof( *mnt_points));

    for ( i = 0; i < mnt_conf_len; ++i) {
        for ( j = 0; j < mnt_proc_len; ++j) {
            if ( strcmp( mnt_conf[ i], mnt_proc[ j]) == 0) {
                do_strdup( mnt_points + mnt_points_len, mnt_conf[ i]);

                ++mnt_points_len;
                mnt_points = try_realloc( mnt_points, mnt_points_len, &max,
                                          MOUNT_BUF_LEN, sizeof( *mnt_points));
                break;
            }
        }
    }

    do_free_double( mnt_proc, mnt_proc_len);
    do_free_double( mnt_conf, mnt_conf_len);
}

#define UMNT_REVERSE 1
#define UMNT_NORMAL 0
/* Unmounts a list of filesystems, either in normal or reverse order. */
static void umount_filesystems( char **mnt, unsigned len, unsigned reverse) {
    int i;
    for ( i = 0; i < len; ++i) {
        do_umount( mnt[ ( reverse ? ( len - i - 1) : i)]);
    }
}


/*
 * Set up functions. Called when getting ready to run a PrivExec
 * application. These should be called in a specific order (top to bottom as
 * listed here) to function correctly (see main.c), and many modify global
 * variables, too. They are broken into functions mostly for better readability.
 */

/* Call this first. */
void init() {

    srand( time( NULL));

    mnt_points_len = 0;
    mnt_targets_len = 0;
    mnt_sys_len = 0;

    tmp_dir = NULL;
    storage_dir = NULL;
    union_dir = NULL;
    dec_dirs_len = 0;
    enc_dirs_len = 0;
    dirs = NULL;

    /*
     * Read from config what needs to mounted and made secure, or available at
     * all, to the PrivExec program.
     */
    get_mount_points();
}

/* Generates directory paths and random names for this session of PrivExec. */
void generate_dir_names() {

    int i;

    /* Tmp dir */
    do_strdup( &tmp_dir, TMP_DIR);
    do_strcat( &tmp_dir, "/");
    do_strcat( &tmp_dir, PRIVEXEC_DIR);
    do_strcat( &tmp_dir, "/");

    /* Storage dir */
    do_strdup( &storage_dir, STORAGE_DIR);
    do_strcat( &storage_dir, "/");
    do_strcat( &storage_dir, PRIVEXEC_DIR);
    do_strcat( &storage_dir, "/");

    /* Overlay dir. */
    do_strdup( &union_dir, tmp_dir);
    do_strcat( &union_dir, RANDOM_NAME);
#ifdef PRIVEXEC_DEBUG
    do_strcat( &union_dir, "_uni"); /* Some readability. */
#endif

    /* Secure container dirs. */
    dec_dirs = malloc( mnt_points_len * sizeof( *dec_dirs));
    enc_dirs = malloc( mnt_points_len * sizeof( *enc_dirs));
    for ( i = 0; i < mnt_points_len; ++i) {
        do_strdup( dec_dirs + i, tmp_dir);
        do_strcat( dec_dirs + i, RANDOM_NAME);
#ifdef PRIVEXEC_DEBUG
        do_strcat( dec_dirs + i, "_dec");
#endif
        ++dec_dirs_len;

        do_strdup( enc_dirs + i, storage_dir);
        do_strcat( enc_dirs + i, RANDOM_NAME);
#ifdef PRIVEXEC_DEBUG
        do_strcat( enc_dirs + i, "_enc");
#endif
        ++enc_dirs_len;
    }
}

/* Creates all the directories used by PrivExec. */
void create_dirs() {

    int i;

    /*
     * Save dirs here as they are created, so that we can remove them later. In
     * case of error, we never touch anything that we did not create.
     */
    dirs = calloc( ( ( mnt_points_len * 2) + 1 + 1), sizeof( *dirs));

    /*
     * These two -- storage and tmp dirs -- are exceptions. They are needed
     * every time, so they won't ever be removed.
     */
    if ( mkdir( storage_dir, S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH))
        if (errno != EEXIST)
            die( "Cannot create storage directory, \"%s\"\n", storage_dir);

    if ( mkdir( tmp_dir, S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH))
        if (errno != EEXIST)
            die( "Cannot create temporary directory, \"%s\"\n", tmp_dir);

    /* Union dir, the fake root. */
    if ( mkdir( union_dir, S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH))
        die( "Cannot create union directory, \"%s\"\n", union_dir);
    dirs[ 0] = union_dir;

    /* Dirs for secure containers. */
    for ( i = 0; i < mnt_points_len; ++i) {

        if ( mkdir( dec_dirs[ i], S_IRUSR | S_IXUSR))
            die( "Cannot create decrypted container directory, \"%s\"\n", dec_dirs[ i]);
        dirs[ 1 + ( i * 2)] = dec_dirs[ i];

        if ( mkdir( enc_dirs[ i], S_IRWXU))
            die( "Cannot create encrypted container directory, \"%s\"\n", enc_dirs[ i]);
        dirs[ 1 + ( i * 2) + 1] = enc_dirs[ i];
    }
}

/*
 * Inserts a dummy ecryptfs key into the keyring that is used for mounting the
 * secure containers, because ecryptfs kernel module expects to see one. This
 * key will never be used for encryption, PrivExec has its own per-process keys.
 */
void insert_key() {

    /* Max size 64. */
    char *pass = RANDOM_NAME; /* We don't really care... */
    char salt[ ECRYPTFS_SALT_SIZE]; /* ...use whatever's in memory. */

    if ( ecryptfs_add_passphrase_key_to_keyring( dummy_key, pass, salt))
        die( "Cannot insert dummy key.\n");

    dummy_key[ ECRYPTFS_SIG_SIZE_HEX] = '\0';
}

/*
 * Forks with the PrivExec clone flag. The parent never returns, so, in effect,
 * the caller becomes a PrivExec program.
 */
void fork_privexec() {
    if ( syscall( SYS_clone, SIGCHLD | CLONE_PRIVEXEC, 0, 0, NULL, NULL)) {
        wait( NULL);
        exit( EXIT_PARENT);
    }
}

/* Mounts a secure container for every private partition. */
void mount_containers() {

    char *ecryptfs_opts;
    int i;

    do_strdup( &ecryptfs_opts, ECRYPTFS_OPTS);
    do_strcat( &ecryptfs_opts, ",ecryptfs_sig=");
    do_strcat( &ecryptfs_opts, dummy_key);
    do_strcat( &ecryptfs_opts, ",ecryptfs_fnek_sig=");
    do_strcat( &ecryptfs_opts, dummy_key);

    for ( i = 0; i < mnt_points_len; ++i) {
        do_mount( enc_dirs[ i], dec_dirs [i],
                  "ecryptfs", MS_NOATIME, ecryptfs_opts);
    }
    free( ecryptfs_opts);
}

/* Mounts an overlay for every private partition. */
void mount_overlays() {

    char *overlayfs_opts;
    int i;

    mnt_targets = malloc( mnt_points_len * sizeof( *mnt_targets));
    mnt_targets_len = 0;

    for ( i = 0; i < mnt_points_len; ++i) {
        do_strdup( mnt_targets + i, union_dir);
        do_strcat( mnt_targets + i, mnt_points[ i]);

        do_strdup( &overlayfs_opts, "lowerdir=");
        do_strcat( &overlayfs_opts, mnt_points[ i]);
        do_strcat( &overlayfs_opts, ",upperdir=");
        do_strcat( &overlayfs_opts, dec_dirs[ i]);

        do_mount( "overlayfs", mnt_targets[ i],
                  "overlayfs", MS_NOATIME, overlayfs_opts);

        ++mnt_targets_len;
        free( overlayfs_opts);
    }
    do_free_double( mnt_points, mnt_points_len);
    mnt_points_len = 0;
}

/*
 * Bind-mounts special filesystems (and everything else) specified in the
 * PrivExec config file. These won't be made private, they are things like /proc
 * and /dev that are necessary to set up a working chroot jail.
 */
void mount_sys() {

    char **mnt;
    unsigned len;
    int i;

    /* Read the config, parse the mount points... */
    mnt = read_mounts_from_conf( &len, PRIVEXEC_SYS_MNT_CONF);
    mnt_sys = malloc( len * sizeof( *mnt));
    mnt_sys_len = 0;

    for ( i = 0; i < len; ++i) {
        do_strdup( mnt_sys + i, union_dir);
        do_strcat( mnt_sys + i, mnt[ i]);

        /* ...and mount them. */
        do_mount( mnt[ i], mnt_sys[ i], NULL, MS_BIND, NULL);

        ++mnt_sys_len;
    }
    do_free_double( mnt, len);
}

/*
 * Final step before executing a private application. Chroots into PrivExec jail
 * and executes the program.
 */
void run_privexec( char **exe) {

    if( fork()) {
        /*
         * Parent will wait until PrivExec application exits, then reverse all
         * the changes to the system we've made so far.
         */
        wait( NULL);
    }
    else {

        int i;
        char *cwd = getcwd( NULL, 0);
        if( !cwd) {
            print_error( "Cannot get current working directory.\n");
            exit( EXIT_FAILURE);
        }

        raise_privileges();

        /* Some permission corrections on the overlay mounts. */
        for ( i = 0; i < mnt_targets_len; ++i) {
            chown( mnt_targets[ i], 0, 0);
            chmod( mnt_targets[ i], S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);
        }

        if ( chdir( union_dir)) {
            print_error( "Cannot change dir to overlay %s.\n", union_dir);
            exit( EXIT_FAILURE);
        }

        /* Go to jail. */
        if ( chroot( union_dir) == -1) {
            print_error( "Cannot chroot to overlay %s.\n", union_dir);
            exit( EXIT_FAILURE);
        }

        /* Drop everything. */
        drop_sudo_privileges();
        drop_privileges();

        /* We must run the application in the original working dir. */
        if ( chdir( cwd)) {
            print_error( "Cannot change dir to %s.\n", cwd);
            exit( EXIT_FAILURE);
        }
        free( cwd);

        /* PrivExec! */
        execvp( *exe, exe);
    }
}


/*
 * Clean up functions. Called when a PrivExec application exits, or for rollback
 * after an error during the launch.
 */

/*
 * Deletes contents of encrypted containers. Not for security, but to save
 * space.
 */
void delete_enc_dirs() {

    int i;
    for ( i = 0; i < enc_dirs_len; ++i)
        do_remove_dir_tree( enc_dirs[ i]);
}

/* Unmounts special filesystems, or any other bind-mounted fs. */
inline void umount_sys() {
    umount_filesystems( mnt_sys, mnt_sys_len, UMNT_REVERSE);
}

/* Unmounts overlays. */
inline void umount_overlays() {
    umount_filesystems( mnt_targets, mnt_targets_len, UMNT_REVERSE);
}

/* Unmounts secure containers. */
inline void umount_containers() {
    umount_filesystems( dec_dirs, dec_dirs_len, UMNT_NORMAL);
}

/* Removes the dummy ecryptfs key from the keyring. */
inline void remove_key() {
    ecryptfs_remove_auth_tok_from_keyring( dummy_key);
}

/* Rmdirs directories created by PrivExec. */
inline void remove_dirs() {
    do_remove_dirs( dirs);
}

/*
 * Absolutely unnecessary since calling this means we are going to die in a
 * moment. These variables are freed as soon as not needed during execution
 * anyway.
 */
void free_memory() {

    do_free_double( mnt_points, mnt_points_len);
    do_free_double( mnt_targets, mnt_targets_len);
    do_free_double( mnt_sys, mnt_sys_len);
    do_free( tmp_dir);
    do_free( storage_dir);
    do_free( union_dir);
    do_free_double( dec_dirs, dec_dirs_len);
    do_free_double( enc_dirs, enc_dirs_len);
    do_free( dirs);
}
